#include <qapplication.h>
#include <qpainter.h>
#include <qbitmap.h>
#include <qimage.h>
#include <qfont.h>

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>

#define IMG_W (256*3)
#define IMG_H (125*3)

//#define IMG_DST_W (256)
//#define IMG_DST_H (125)

#define IMG_USAGE (0.75f)

static const char font_name[] = "FreeSerif";
static char allowed_chars[] = "0123456789";
static const int n_chars = 6;

static int n_images = 0;


inline float rand(float min, float max)
{
    return (rand()/(float)RAND_MAX) * (max-min) + min;
}

inline float clamp(float v, float min, float max)
{
    if ( v < min ) return min;
    else if ( v > max ) return max;
    else return v;
}
QImage apply_lens( const QImage& img, int lx, int ly, int r )
{
  int w = img.width();
  int h = img.height();
  QImage res = QImage( w,h, img.depth(), img.numColors() );

  for ( int y=0; y<h; ++y )
  {
      for ( int x=0; x<w; ++x )
      {
          int xx = x, yy = y;
          int dx = (x-lx), dy = (y-ly);

          float d = sqrt(dx*dx + dy*dy);
          float new_d = (d-r)>0 ? (d-r) : 0;

          if ( d > 0.f )
          {
              xx = (int)clamp(lx + dx/d * new_d, 0, w-1);
              yy = (int)clamp(ly + dy/d * new_d, 0, h-1);
          }
          res.setPixel(x,y, img.pixel(xx,yy));
      }
  }
  return res;
}

inline bool Is_Black( const QRgb& color  )
{
    return qGreen(color) < 128;
}

bool Has_Black_On_Borders( const QImage& img )
{
    int w = img.width(), h = img.height();

    for ( int y=0; y<IMG_H; ++y )
        if ( Is_Black(img.pixel(0,y)) ||
             Is_Black(img.pixel(w-1,y)) )
                 return true;

    for ( int x=0; x<IMG_W; ++x )
        if ( Is_Black(img.pixel(x,0)) ||
             Is_Black(img.pixel(x,h-1)) )
                 return true;

    return false;
}

void Print_Usage_And_Exit( const char* progname )
{
    printf( "Usage: %s <num>\n...where <num> is the amount of pictures to be generated.\n\n", progname );
    exit(2);
}

int main( int argc, char **argv )
{
    if ( argc < 2 )
        Print_Usage_And_Exit(argv[0]);

    // See how many files to generate
    bool toint_ok = false;
    n_images = QString(argv[1]).toInt( &toint_ok );
    if ( !toint_ok || n_images < 1 )
        Print_Usage_And_Exit(argv[0]);

    puts("Initializing.");

    srand(time(0));
    QApplication app(argc, argv, true);

    for ( int img_i=0; img_i<n_images; ++img_i )
    {
        printf("Generating image %d of %d...\n", (img_i+1), n_images);

        QPixmap pixmap(IMG_W, IMG_H);
        QPainter painter( &pixmap );
        painter.setFont(QFont(font_name, 16));

        // Clear the bitmap
        painter.fillRect ( 0,0, IMG_W,IMG_H, QBrush(Qt::white));
        painter.setBrush(QBrush(Qt::black));

        // Randomize the text
        QString text("");
        int n_allowed_chars = strlen(allowed_chars);
        for ( int i=0; i<n_chars; ++i )
            text.append(allowed_chars[(int)rand(0,n_allowed_chars)]);

        QFont font( painter.font());
        font.setStyleStrategy( QFont::NoAntialias );

        // Find a good font size (one that spans about IMG_USAGE of the
        // image horizontally) using binary search:
        float fontsize_min = 1, fontsize_max = 250, last_w_diff = -1;
        for ( int i=0; i<100; ++i )
        {
            font.setPointSizeFloat( (fontsize_min + fontsize_max)/2.f );
            QFontMetrics fm( font );

            float w_diff = (float)fm.width(text) - (IMG_W * IMG_USAGE);
            float new_lim = (fontsize_min + fontsize_max)/2.f;

            if ( (w_diff <= 10 && fm.height() < IMG_H) || last_w_diff == w_diff )
                break;

            if ( w_diff < 0 )
                fontsize_min = new_lim;
            else if ( w_diff > 0 )
                fontsize_max = new_lim;

            last_w_diff = w_diff;
        }
        painter.setFont( font );

        // Draw the text
        QFontMetrics fm( font );
        painter.drawText( (IMG_W-fm.width(text))/2, (IMG_H+fm.ascent())/2, text );
        painter.end();

        QImage img = pixmap.convertToImage();

        if ( Has_Black_On_Borders( img ))
        {
            puts("BUG: rendered text was larger than the image.");
            return 0;
        }

        // Distort the text by applying some lens effects
        float margin = IMG_W*(1.f-IMG_USAGE)/2;
        for ( int i=0,max=12; i<5 && max>0; ++i, --max )
        {
            int lx = (int)rand(margin*0.5f, IMG_W - (margin*0.5f));
            int ly = (i&1)
                ? (int)rand(0, IMG_H/2) // upper half
                : (int)rand(IMG_H/2, IMG_H); // lower half

            // Don't stretch black areas (compress only)
            if ( Is_Black(img.pixel(lx,ly)))
            { --i; continue; }

            int r = (int)rand(IMG_W*(1.f-IMG_USAGE)*0.1f, IMG_W*(1.f-IMG_USAGE)*0.3f);
            QImage img2 = apply_lens( img, lx,ly, r);

            // Don't let the text go out of the image
            if ( Has_Black_On_Borders( img2 ))
            { --i; continue; }

            img = img2;
        }

        // Add some negating "pies" to further clutter the text
        pixmap = img;
        painter.begin( &pixmap );
        painter.setRasterOp( Qt::NotROP );
        painter.setBrush(QBrush(Qt::black));
        painter.setPen(QPen(Qt::NoPen));
        for (int i=0; i<3; ++i )
        {
            int ang = (int)rand(16*360*0.2, 16*360*0.8);
            painter.drawPie(
                (int)rand(0, IMG_W),
                (int)rand(0, IMG_H),
                (int)rand(IMG_W*0.2, IMG_W),
                (int)rand(IMG_H*0.2, IMG_H),
                ang, (int)rand(ang+(16*360-ang)*0.1f, 16*360) );
        }
        painter.end();

        // Save the image
        img = pixmap.convertToImage();
        //img = img.smoothScale( IMG_DST_W, IMG_DST_H );
        img = img.convertDepth( 1, Qt::ThresholdAlphaDither );
        pixmap = img;
        pixmap.save ( text + ".png", "PNG" );
    }
    return 0;
}
